package de.unikoblenz.fgbks.base.utils.boundary;

import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.INCLUSIVE;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.APPEND;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.BETWEEN;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.COMBINE;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.INTERSECTION;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.LOWER_BOUNDS;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.UPPER_BOUNDS;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_EQUAL;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_IN_CONTACT;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_NOT_IN_CONTACT;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_OVERLAPPING;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_SUBSUMPTION;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.SUBSUMES;
import static de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType.LOWER;
import static de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType.UPPER;

import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterAppend;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterBetween;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterCombine;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterIntersection;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterLowerBounds;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterUpperBounds;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreater;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheck;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckEqual;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckIsSubsumption;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckNotInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckOverlap;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckSubsumes;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreater;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.creater.CreaterLower;
import de.unikoblenz.fgbks.base.utils.boundary.creater.CreaterUpper;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import org.apache.commons.lang3.Validate;

/**
 * This abstract boundary is for growing boundaries, like numbers or dates.
 *
 * @param <T> the type of the boundary
 */
public abstract class AbstractGrowingBoundary<T extends Comparable<? super T>>
    extends AbstractBoundary<T> {

  private static HashMap<BoundaryCheckType, BoundaryCheck> checkerMap = new HashMap<>();
  private static HashMap<BoundaryCreaterType, BoundaryCreater> createrMap = new HashMap<>();
  private static HashMap<BoundaryBiCreaterType, BoundaryBiCreater> biCreaterMap = new HashMap<>();

  static {
    checkerMap.put(IS_EQUAL, CheckEqual.getInstance());
    checkerMap.put(IS_IN_CONTACT, CheckInContact.getInstance());
    checkerMap.put(IS_NOT_IN_CONTACT, CheckNotInContact.getInstance());
    checkerMap.put(IS_OVERLAPPING, CheckOverlap.getInstance());
    checkerMap.put(IS_SUBSUMPTION, CheckIsSubsumption.getInstance());
    checkerMap.put(SUBSUMES, CheckSubsumes.getInstance());
    createrMap.put(UPPER, CreaterUpper.getInstance());
    createrMap.put(LOWER, CreaterLower.getInstance());
    biCreaterMap.put(INTERSECTION, BiCreaterIntersection.getInstance());
    biCreaterMap.put(LOWER_BOUNDS, BiCreaterLowerBounds.getInstance());
    biCreaterMap.put(UPPER_BOUNDS, BiCreaterUpperBounds.getInstance());
    biCreaterMap.put(BETWEEN, BiCreaterBetween.getInstance());
    biCreaterMap.put(APPEND, BiCreaterAppend.getInstance());
    biCreaterMap.put(COMBINE, BiCreaterCombine.getInstance());
  }

  protected T lowerBound;
  protected BoundType lowerBoundType;
  protected T upperBound;
  protected BoundType upperBoundType;

  /**
   * Get the current lower bound of the boundary.
   *
   * @return the lower bound
   */
  public T getLowerBound() {
    return lowerBound;
  }

  /**
   * Get the current upper bound of the boundary.
   *
   * @return the lower bound
   */
  public T getUpperBound() {
    return upperBound;
  }

  /**
   * Get the {@link BoundType} of the lower bound.
   *
   * @return the lower {@link BoundType}
   */
  public BoundType getLowerBoundType() {
    return lowerBoundType;
  }

  /**
   * Get the {@link BoundType} of the upper bound.
   *
   * @return the upper {@link BoundType}
   */
  public BoundType getUpperBoundType() {
    return upperBoundType;
  }

  protected AbstractGrowingBoundary(String text, Class<T> type) {
    super(text, type);
  }

  protected AbstractGrowingBoundary(Class<T> type) {
    super(type);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected void validate() throws BoundaryParseException {
    Validate.notNull(lowerBound);
    Validate.notNull(upperBound);
    if (lowerBound.equals(getMinValue())) {
      lowerBoundType = INCLUSIVE;
    }
    if (upperBound.equals(getMaxValue())) {
      upperBoundType = INCLUSIVE;
    }
    Validate.notNull(lowerBoundType);
    Validate.notNull(upperBoundType);
    if (lowerBound.compareTo(upperBound) > 0) {
      throw new BoundaryParseException(
          String.format("Lower bound %s is greater than upper bound %s.", lowerBound, upperBound));
    }
    // if equal -> both must be INCLUSIVE!
    if (upperBound.equals(lowerBound)
        && (lowerBoundType != INCLUSIVE || upperBoundType != INCLUSIVE)) {
      throw new BoundaryParseException("If bounds are equal -> both bounds must be INCLUSIVE!");
    }
  }

  @Override
  protected Map<BoundaryCheckType, BoundaryCheck> getChecker() {
    return Collections.unmodifiableMap(checkerMap);
  }

  @Override
  protected Map<BoundaryBiCreaterType, BoundaryBiCreater> getBiCreater() {
    return Collections.unmodifiableMap(biCreaterMap);
  }

  @Override
  protected Map<BoundaryCreaterType, BoundaryCreater> getCreater() {
    return Collections.unmodifiableMap(createrMap);
  }

  @Override
  public int hashCode() {
    return Objects.hash(lowerBound, lowerBoundType, upperBound, upperBoundType);
  }

  @Override
  public int compareTo(Boundary o) {
    if (!(o instanceof AbstractGrowingBoundary)) {
      return 1;
    }
    int comLower = compareToLowerBound((AbstractGrowingBoundary) o);
    return comLower == 0 ? compareToUpperBound((AbstractGrowingBoundary) o) : comLower;
  }

  /**
   * Compare the lower bounds.
   *
   * @param o other boundary
   * @return -1 if current lower bound is lower, 0 if equal, 1 if higher
   */
  public int compareToLowerBound(AbstractGrowingBoundary o) {
    if (lowerBound.equals(o.lowerBound)) {
      if (lowerBoundType.equals(o.lowerBoundType)) {
        return 0;
      }
      return lowerBoundType == INCLUSIVE ? -1 : 1;
    } else if (lowerBound.equals(getMinValue())) {
      return -1;
    } else if (o.lowerBound.equals(getMinValue())) {
      return 1;
    }
    return lowerBound.compareTo((T) o.lowerBound);
  }

  /**
   * Compare the upper bounds.
   *
   * @param o other boundary
   * @return -1 if current upper bound is lower, 0 if equal, 1 if higher
   */
  public int compareToUpperBound(AbstractGrowingBoundary o) {
    if (upperBound.equals(o.upperBound)) {
      if (upperBoundType.equals(o.upperBoundType)) {
        return 0;
      }
      return upperBoundType == INCLUSIVE ? 1 : -1;
    } else if (upperBound.equals(getMaxValue())) {
      return 1;
    } else if (o.upperBound.equals(o.getMaxValue())) {
      return -1;
    }
    return upperBound.compareTo((T) o.upperBound);
  }

  /**
   * Get the minimum possible value of the boundary (-INFINITY).
   *
   * @return the minimum value
   */
  public abstract T getMinValue();

  /**
   * Get the maximum possible value of the boundary (INFINITY).
   *
   * @return the maximum value
   */
  public abstract T getMaxValue();

  /**
   * Get a copy of the boundary.
   *
   * @param boundary the original boundary
   * @return a copy of the given boundary
   */
  public AbstractGrowingBoundary<T> getCopy(AbstractGrowingBoundary<T> boundary) {
    return getCopy(
        boundary.lowerBound, boundary.upperBound, boundary.lowerBoundType, boundary.upperBoundType);
  }

  /**
   * Get a copy of the boundary.
   *
   * @param lowerBound the lower bound
   * @param upperBound the upper bound
   * @param lowerBoundType the lower {@link BoundType}
   * @param upperBoundType the upper {@link BoundType}
   * @return a copy of the given boundary
   */
  public abstract AbstractGrowingBoundary<T> getCopy(
      T lowerBound, T upperBound, BoundType lowerBoundType, BoundType upperBoundType);

  /**
   * The structure of the parsed text of the boundary looks like <br>
   *
   * <ul>
   *   <li>"= 10"<br>
   *       {lowerBound=10; upperBound=10; lowerBoundType={@link BoundType#INCLUSIVE},
   *       upperBoundType={@link BoundType#INCLUSIVE})
   *   <li>"{@literal <}= 10"<br>
   *       {lowerBound=-INFINITY; upperBound=10; lowerBoundType={@link BoundType#INCLUSIVE},
   *       upperBoundType={@link BoundType#INCLUSIVE})
   *   <li>"{@literal <}10"<br>
   *       {lowerBound=-INFINITY; upperBound=10; lowerBoundType={@link BoundType#INCLUSIVE},
   *       upperBoundType={@link BoundType#EXCLUSIVE})
   *   <li>"{@literal >} 10"<br>
   *       {lowerBound=10; upperBound=INFINITY; lowerBoundType={@link BoundType#EXCLUSIVE},
   *       upperBoundType={@link BoundType#INCLUSIVE})
   *   <li>"{@literal >}= 10"<br>
   *       {lowerBound=10; upperBound=INFINITY; lowerBoundType={@link BoundType#INCLUSIVE},
   *       upperBoundType={@link BoundType#INCLUSIVE})
   *   <li>"[10..20]" <br>
   *       {lowerBound=10; upperBound=20; lowerBoundType={@link BoundType#INCLUSIVE},
   *       upperBoundType={@link BoundType#INCLUSIVE})
   *   <li>"]10..20]" <br>
   *       {lowerBound=10; upperBound=20; lowerBoundType={@link BoundType#EXCLUSIVE},
   *       upperBoundType={@link BoundType#INCLUSIVE})
   *   <li>"]10..20[" <br>
   *       {lowerBound=10; upperBound=20; lowerBoundType={@link BoundType#EXCLUSIVE},
   *       upperBoundType={@link BoundType#EXCLUSIVE})
   *   <li>"[10..20[" <br>
   *       {lowerBound=10; upperBound=20; lowerBoundType={@link BoundType#INCLUSIVE},
   *       upperBoundType={@link BoundType#EXCLUSIVE})
   *   <li>"" (empty String) <br>
   *       {lowerBound=-INFINITY; upperBound=INFINITY; lowerBoundType={@link BoundType#INCLUSIVE},
   *       upperBoundType={@link BoundType#INCLUSIVE})
   * </ul>
   *
   * @return the parsed text as string
   */
  @Override
  public String getParsedText() {
    if (lowerBound.equals(getMinValue()) && upperBound.equals(getMaxValue())) {
      return "";
    }
    if (lowerBound.equals(upperBound)) {
      return "= " + lowerBound;
    }
    StringBuilder sb = new StringBuilder();
    if (lowerBound.equals(getMinValue())) {
      sb.append("<");
      if (getUpperBoundType() == INCLUSIVE) {
        sb.append("=");
      }
      sb.append(" ");
      sb.append(getUpperBound());
    } else if (upperBound.equals(getMaxValue())) {
      sb.append(">");
      if (getLowerBoundType() == INCLUSIVE) {
        sb.append("=");
      }
      sb.append(" ");
      sb.append(getLowerBound());
    } else {
      sb.append(getLowerBoundType() == INCLUSIVE ? "[" : "]");
      sb.append(getLowerBound());
      sb.append("..");
      sb.append(getUpperBound());
      sb.append(getUpperBoundType() == INCLUSIVE ? "]" : "[");
    }
    return sb.toString();
  }
}
