package de.unikoblenz.fgbks.base.utils.boundary.impl;

import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.EXCLUSIVE;
import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.INCLUSIVE;

import de.unikoblenz.fgbks.base.utils.boundary.AbstractGrowingBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.BoundType;
import de.unikoblenz.fgbks.base.utils.boundary.BoundaryParseException;
import org.apache.commons.lang3.Validate;

/**
 * The long boundary is a rage between two long values. And looks as string representation like
 *
 * <ul>
 *   <li>"= 10"<br>
 *       {lowerBound=10; upperBound=10; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"{@literal <}= 10"<br>
 *       {lowerBound=-INFINITY; upperBound=10; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"{@literal <} 10"<br>
 *       {lowerBound=-INFINITY; upperBound=10; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#EXCLUSIVE})
 *   <li>"{@literal >} 10"<br>
 *       {lowerBound=10; upperBound=INFINITY; lowerBoundType={@link BoundType#EXCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"{@literal >}= 10"<br>
 *       {lowerBound=10; upperBound=INFINITY; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"[10..20]" <br>
 *       {lowerBound=10; upperBound=20; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"]10..20]" <br>
 *       {lowerBound=10; upperBound=20; lowerBoundType={@link BoundType#EXCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"]10..20[" <br>
 *       {lowerBound=10; upperBound=20; lowerBoundType={@link BoundType#EXCLUSIVE},
 *       upperBoundType={@link BoundType#EXCLUSIVE})
 *   <li>"[10..20[" <br>
 *       {lowerBound=10; upperBound=20; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#EXCLUSIVE})
 *   <li>"" (empty String) <br>
 *       {lowerBound=-INFINITY; upperBound=INFINITY; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 * </ul>
 */
public class LongBoundary extends AbstractGrowingBoundary<Long> {

  /**
   * Parse the input text to a new Long boundary. See {@link LongBoundary} for the required
   * structure.
   *
   * @param text the text, which will be parsed to the new LongBoundary object
   */
  public LongBoundary(String text) {
    super(text, Long.TYPE);
  }

  private LongBoundary() {
    super(Long.TYPE);
  }

  @Override
  public AbstractGrowingBoundary<Long> getCopy(
      Long lowerBound, Long upperBound, BoundType lowerBoundType, BoundType upperBoundType) {
    LongBoundary v = new LongBoundary();
    v.lowerBound = lowerBound;
    v.upperBound = upperBound;
    v.lowerBoundType = lowerBoundType;
    v.upperBoundType = upperBoundType;
    v.simplifyBound();
    v.validate();
    return v;
  }

  @Override
  public Long getMinValue() {
    return Long.MIN_VALUE;
  }

  @Override
  public Long getMaxValue() {
    return Long.MAX_VALUE;
  }

  @Override
  protected void parse(String text) {
    try {
      text = Validate.notNull(text).trim();
      if (text.isEmpty()) {
        upperBound = getMaxValue();
        lowerBound = getMinValue();
        lowerBoundType = INCLUSIVE;
        upperBoundType = INCLUSIVE;
      } else if (text.startsWith("[") || text.startsWith("]")) {
        lowerBoundType = text.startsWith("[") ? INCLUSIVE : EXCLUSIVE;
        upperBoundType = text.endsWith("]") ? INCLUSIVE : EXCLUSIVE;
        text = text.replace('[', '\0').replace(']', '\0').trim();
        String[] n = text.split("\\.\\.");
        lowerBound = Long.parseLong(n[0]);
        upperBound = Long.parseLong(n[1]);
      } else if (text.startsWith(">")) {
        lowerBoundType = text.contains("=") ? INCLUSIVE : EXCLUSIVE;
        text = text.replace('=', '\0').replace('>', '\0').trim();
        lowerBound = Long.parseLong(text);
        upperBound = getMaxValue();
        upperBoundType = INCLUSIVE;
      } else if (text.startsWith("<")) {
        upperBoundType = text.contains("=") ? INCLUSIVE : EXCLUSIVE;
        text = text.replace('=', '\0').replace('<', '\0').trim();
        upperBound = Long.parseLong(text);
        lowerBound = getMinValue();
        lowerBoundType = INCLUSIVE;
      } else if (text.startsWith("=")) {
        text = text.replace('=', '\0').trim();
        upperBound = Long.parseLong(text);
        lowerBound = Long.parseLong(text);
        lowerBoundType = INCLUSIVE;
        upperBoundType = INCLUSIVE;
      } else {
        throw new BoundaryParseException();
      }
    } catch (Exception e) {
      throw new BoundaryParseException(
          String.format(
              "Parsing Long boundary failed: %s. Exception: %s", this.getText(), e.getMessage()));
    }
    simplifyBound();
  }

  private void simplifyBound() {
    if (lowerBoundType == EXCLUSIVE) {
      lowerBound = lowerBound + 1;
      lowerBoundType = INCLUSIVE;
    }
    if (upperBoundType == EXCLUSIVE) {
      upperBound = upperBound - 1;
      upperBoundType = INCLUSIVE;
    }
  }
}
