package de.unikoblenz.fgbks.base.value;

import static org.apache.commons.lang3.Validate.inclusiveBetween;
import static org.apache.commons.lang3.Validate.matchesPattern;
import static org.apache.commons.lang3.Validate.notBlank;
import static org.apache.commons.lang3.Validate.notNull;

public abstract class AbstractStringValueObject extends AbstractSimpleValueObject<String> {

  public static final int DEFAULT_MIN_LEN = 1;

  protected AbstractStringValueObject() {}

  public AbstractStringValueObject(String initialValue) {
    super(initialValue);
  }

  @Override
  public String toString() {
    return getValue();
  }

  @Override
  public String getValue() {
    if (handleBlanksAsNull() && super.getValue().trim().isEmpty()) {
      return null;
    }
    return isBlankAllowed() ? super.getValue() : toStringTrim();
  }

  private String toStringTrim() {
    return super.getValue().trim();
  }

  protected String getClassName() {
    return getClass().getSimpleName();
  }

  protected boolean handleBlanksAsNull() {
    return true;
  }

  @Override
  protected String validateAndNormalize(String newValue) {
    String retValue = super.validateAndNormalize(newValue);
    validateBlank(retValue);
    validateLength(retValue);
    validatePattern(retValue);
    return retValue;
  }

  protected String getValidationPattern() {
    return null;
  }

  protected Integer getMinLength() {
    return DEFAULT_MIN_LEN;
  }

  protected abstract Integer getMaxLength();

  protected boolean isBlankAllowed() {
    return getMinLength() == 0;
  }

  protected String getNumbersOnlyRegex(int numDigits) {
    return String.format("[0-9]{0,%d}", numDigits);
  }

  private void validatePattern(String newValue) {
    String validationPattern = getValidationPattern();
    if (null != validationPattern) {
      matchesPattern(
          newValue,
          validationPattern,
          "%s is not a valid %s %s",
          newValue,
          getClassName(),
          validationPattern);
    }
  }

  private void validateLength(String newValue) {
    Integer min = getMinLength();
    Integer max = getMaxLength();
    inclusiveBetween(
        min,
        max,
        newValue.length(),
        "%s is not a valid %s. Its length must be between %d and %d.",
        newValue,
        getClassName(),
        min,
        max);
  }

  private void validateBlank(String retValue) {
    if (!isBlankAllowed()) {
      notNull(retValue, "%s can't be null", getClassName());
      notBlank(retValue, "%s can't be blank", getClassName());
    }
  }
}
