package de.unikoblenz.fgbks.core.dmn.domain.vdmn.utils;

import de.unikoblenz.fgbks.base.domain.Name;
import de.unikoblenz.fgbks.core.dmn.domain.ids.RuleId;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnColumn;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecision;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecisionTable;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDefinition;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputColumn;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputValue;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnOutputColumn;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnRule;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnValue;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VTypeRef;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;
import org.apache.commons.lang3.Validate;

/**
 * VDmnFunctions is a utility class, providing functions for the VDmn classes.
 */
public class VDmnFunctions {

  private VDmnFunctions() {}
  /**
   * Generate a list of {@link VDmnColumn}, which are in the {@link VDmnDefinition} and has the
   * defined type ({@link VTypeRef}).
   *
   * @param dmnDefinition the {@link VDmnDefinition}
   * @param vTypeRef the type as {@link VTypeRef}
   * @return a new list of {@link VDmnColumn}s
   */
  public static List<VDmnColumn> getColumnsByTypeRef(
      VDmnDefinition dmnDefinition, VTypeRef vTypeRef) {
    Validate.notNull(dmnDefinition);
    Validate.notNull(vTypeRef);
    List<VDmnColumn> columns = new ArrayList<>();
    columns.addAll(
        dmnDefinition.getDmnDecisions().stream()
            .flatMap(
                v ->
                    v.getDmnDecisionTable().getInputColumns().stream()
                        .filter(c -> c.getTypeRef() == vTypeRef))
            .collect(Collectors.toList()));
    columns.addAll(
        dmnDefinition.getDmnDecisions().stream()
            .flatMap(
                v ->
                    v.getDmnDecisionTable().getOutputColumns().stream()
                        .filter(c -> c.getTypeRef() == vTypeRef))
            .collect(Collectors.toList()));
    return columns;
  }

  /**
   * Generate a list of all output columns ({@link VDmnOutputColumn}) of the given {@link
   * VDmnDefinition}.
   *
   * @param dmnDefinition the {@link VDmnDefinition}
   * @return a new list with all output columns ({@link VDmnOutputColumn})
   */
  public static List<VDmnOutputColumn> getOutputColumns(VDmnDefinition dmnDefinition) {
    return Validate.notNull(dmnDefinition).getDmnDecisions().stream()
        .flatMap(v -> v.getDmnDecisionTable().getOutputColumns().stream())
        .collect(Collectors.toList());
  }

  /**
   * Generate a list of {@link VDmnValue}s, which are in the specified {@link VDmnColumn} and which
   * are filtered with the given list of {@link VDmnRule}.
   *
   * @param column the {@link VDmnColumn}, containing the {@link VDmnValue}s
   * @param rules the list of {@link VDmnRule}
   * @return a new list of {@link VDmnValue}
   */
  public static List<VDmnValue> getColumnValuesInRules(VDmnColumn column, List<VDmnRule> rules) {
    return getColumnValuesInRuleIds(
        column, rules.stream().map(VDmnRule::getRuleId).collect(Collectors.toList()));
  }

  /**
   * Generate a list of {@link VDmnValue}s, which are in the specified {@link VDmnColumn} and which
   * are filtered with the given list of {@link RuleId}.
   *
   * @param column the {@link VDmnColumn}, containing the {@link VDmnValue}s
   * @param ruleIds the list of {@link RuleId}
   * @return a new list of {@link VDmnValue}
   */
  public static List<VDmnValue> getColumnValuesInRuleIds(VDmnColumn column, List<RuleId> ruleIds) {
    return column.getValues().stream()
        .filter(v -> ruleIds.contains(v.getRuleId()))
        .collect(Collectors.toList());
  }

  /**
   * Generate a list of {@link VDmnInputValue}s, which are in the specified {@link VDmnInputColumn}
   * and which are filtered with the given list of {@link VDmnInputValue}.
   *
   * @param column the {@link VDmnInputColumn}, containing the {@link VDmnInputValue}s
   * @param rules the list of {@link VDmnRule}
   * @return a new list of {@link VDmnInputValue}
   */
  public static List<VDmnInputValue> getColumnValuesInRules(
      VDmnInputColumn column, List<VDmnRule> rules) {
    return getColumnValuesInRuleIds(
        column, rules.stream().map(VDmnRule::getRuleId).collect(Collectors.toList()));
  }

  /**
   * Generate a list of {@link VDmnInputValue}s, which are in the specified {@link VDmnInputColumn}
   * and which are filtered with the given list of {@link RuleId}.
   *
   * @param column the {@link VDmnInputColumn}, containing the {@link VDmnInputValue}s
   * @param ruleIds the list of {@link RuleId}
   * @return a new list of {@link VDmnInputValue}
   */
  public static List<VDmnInputValue> getColumnValuesInRuleIds(
      VDmnInputColumn column, List<RuleId> ruleIds) {
    return column.getInputValues().stream()
        .filter(v -> ruleIds.contains(v.getRuleId()))
        .collect(Collectors.toList());
  }

  /**
   * Check, if the given {@link VDmnRule}s have different conclusions.
   *
   * @param rules the {@link VDmnRule}s to check
   * @return true, if at least one rule has a different conclusion
   */
  public static boolean differentConclusions(List<VDmnRule> rules) {
    if (Validate.notNull(rules).size() == 0) {
      return false;
    }
    VDmnRule refRule = rules.get(0);
    for (VDmnRule rule : rules) {
      if (differentConclusions(refRule, rule)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Check, if the given {@link VDmnRule}s have different conclusions.
   *
   * @param rules the {@link VDmnRule}s to check
   * @return true, if at least one rule has a different conclusion
   */
  public static boolean sameConclusions(List<VDmnRule> rules) {
    return !differentConclusions(rules);
  }

  /**
   * Check, if the given two {@link VDmnRule}s have different conclusions (outputs).
   *
   * @param oneRule the first {@link VDmnRule} to check
   * @param otherRule the second {@link VDmnRule} to check
   * @return true, if both rules has different conclusions.
   */
  public static boolean differentConclusions(VDmnRule oneRule, VDmnRule otherRule) {
    if (Validate.notNull(oneRule).getDmnOutputValues().size()
        != Validate.notNull(otherRule).getDmnOutputValues().size()) {
      return true;
    }
    for (int i = 0; i < oneRule.getDmnOutputValues().size(); i++) {
      if (!oneRule
          .getDmnOutputValues()
          .get(i)
          .getText()
          .equals(otherRule.getDmnOutputValues().get(i).getText())) {
        return true;
      }
    }
    return false;
  }

  /**
   * Check, if the given two {@link VDmnRule}s have the same conclusions (outputs).
   *
   * @param oneRule the first {@link VDmnRule} to check
   * @param otherRule the second {@link VDmnRule} to check
   * @return true, if both rules has the same conclusions.
   */
  public static boolean sameConclusions(VDmnRule oneRule, VDmnRule otherRule) {
    return !differentConclusions(oneRule, otherRule);
  }

  /**
   * Get a nested list of {@link VDmnRule} from the given {@link VDmnDecisionTable}. Each element in
   * the list is a cluster of rules, which has the same conclusion.
   *
   * @param dmnDecisionTable the {@link VDmnDecisionTable}
   * @return the nested list of {@link VDmnRule}
   */
  public static List<List<VDmnRule>> getRuleClustersWithIdenticalOutput(
      VDmnDecisionTable dmnDecisionTable) {
    List<List<VDmnRule>> clusters = new ArrayList<>();
    List<VDmnRule> remainingRules = new ArrayList<>(dmnDecisionTable.getRules());
    for (int i = 0; i < remainingRules.size(); i++) {
      List<VDmnRule> newCluster = new ArrayList<>();
      if (remainingRules.get(i) != null) {
        newCluster.add(remainingRules.get(i));
        remainingRules.set(i, null);
        for (int u = i + 1; u < remainingRules.size(); u++) {
          if (remainingRules.get(u) != null
              && sameConclusions(newCluster.get(0), remainingRules.get(u))) {
            newCluster.add(remainingRules.get(u));
            remainingRules.set(u, null);
          }
        }
        if (newCluster.size() > 1) {
          clusters.add(newCluster);
        }
      }
    }
    return clusters;
  }

  /**
   * Get the name of a column as string. If a label is present, the string of the label is returned.
   * If the label is not present, the name is returned. If the name is not present, "[no name]" is
   * returned.
   *
   * @param column the {@link VDmnColumn}
   * @return a string, representing the "name" of the column
   */
  public static String getColumnStringName(VDmnColumn column) {
    Validate.notNull(column);
    return column.getLabel().isPresent()
        ? column.getLabel().get().toString()
        : column.getName().orElse(Name.NO_NAME).toString();
  }

  /**
   * Get a string with all row numbers of the given list of {@link VDmnRule}. The rows are sorted.
   *
   * @param rules the collection of {@link VDmnRule}
   * @return the string of (sorted) rows
   */
  public static String getRulesRowsStrings(Collection<VDmnRule> rules) {
    return Validate.notNull(rules).stream()
        .map(r -> r.getRowNumber().toString())
        .sorted()
        .collect(Collectors.joining(", "));
  }

  /**
   * Get the String for a decision (for message text)
   *
   * @param decision the {@link VDmnDecision}
   * @return a string "Decision %s: '
   */
  public static String templateDecision(VDmnDecision decision) {
    return String.format("Decision \"%s\": ", decision.getName().orElse(Name.NO_NAME));
  }

  /**
   * Get the String for a decision and a column (for message text)
   *
   * @param column the {@link VDmnColumn}
   * @return a string "Decision %s, Column %s: '
   */
  public static String templateDecisionColumn(VDmnColumn column) {
    return String.format(
        "Decision \"%s\", Column \"%s\": ",
        column.getDmnDecision().getName().orElse(Name.NO_NAME), getColumnStringName(column));
  }

  /**
   * Get the String for a decision and a rule (for message text)
   *
   * @param rule the {@link VDmnRule}
   * @return a string "Decision %s, Rule %s: '
   */
  public static String templateDecisionRule(VDmnRule rule) {
    return String.format(
        "Decision \"%s\", Rule \"%s\": ",
        rule.getDmnDecision().getName().orElse(Name.NO_NAME), rule.getRowNumber());
  }
}
