package de.unikoblenz.fgbks.core.dmn.verification.result;

import de.unikoblenz.fgbks.base.builder.DefaultBuilder;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.Verifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.VerificationType;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import javax.json.bind.annotation.JsonbProperty;
import org.apache.commons.lang3.Validate;

/**
 * A verifier result represents the output of one {@link Verifier#verify(ExecutorService)}. So it
 * contains a {@link VerificationType} and a set of {@link VerificationResultEntry}ies. These are
 * single verification errors, which were determined by the verifier.
 */
public class VerifierResult extends AbstractResultObject {

  private VerificationType verificationType;
  private Long executionTime;
  private Set<VerificationResultEntry> verificationResultEntries;

  /**
   * Create a new VerifierResult.
   */
  protected VerifierResult() {
    this.verificationResultEntries = ConcurrentHashMap.newKeySet();
  }

  /**
   * Get a new instance of a builder for a {@link VerifierResult}.
   *
   * @return the instance of a {@link Builder}
   */
  public static Builder getBuilder() {
    return new VerifierResult().new Builder();
  }

  /**
   * Get the {@link VerificationType} of the {@link Verifier}.
   *
   * @return the {@link VerificationType}
   */
  @JsonbProperty("type")
  public VerificationType getVerificationType() {
    return verificationType;
  }

  /**
   * Get the set of {@link VerificationResultEntry}ies. Each {@link VerificationResultEntry}
   * expresses one verification error.
   *
   * @return the set of {@link VerificationResultEntry}ies
   */
  @JsonbProperty("entries")
  public Set<VerificationResultEntry> getVerificationResultEntries() {
    return new HashSet<>(verificationResultEntries);
  }

  /**
   * Get the amount of {@link VerificationResultEntry}ies.
   *
   * @return the amount of {@link VerificationResultEntry}ies
   */
  @JsonbProperty("size")
  public int getAmountOfEntries() {
    return verificationResultEntries.size();
  }

  /**
   * Get the execution time of the verifier in nano seconds.
   *
   * @return the execution time in nano seconds
   */
  @JsonbProperty("executionTime")
  public Long getExecutionTime() {
    return executionTime;
  }

  /** A builder class for {@link VerifierResult}. */
  public class Builder extends DefaultBuilder<VerifierResult> {

    /**
     * Set the {@link VerificationType} of {@link VerifierResult}.
     *
     * @param verificationType the {@link VerificationType}
     * @return the current builder
     */
    public Builder withVerificationType(VerificationType verificationType) {
      value.verificationType = Validate.notNull(verificationType);
      return this;
    }

    /**
     * Add a new {@link VerificationResultEntry} to the {@link VerifierResult}.
     *
     * @param verificationResultEntry the {@link VerificationResultEntry}
     * @return the current builder
     */
    public Builder addVerificationResultEntry(VerificationResultEntry verificationResultEntry) {
      value.verificationResultEntries.add(Validate.notNull(verificationResultEntry));
      return this;
    }

    /**
     * Set the execution time of {@link VerifierResult}.
     *
     * @param executionTime the execution time of the verifier
     * @return the current builder
     */
    public Builder withExecutionTime(Long executionTime) {
      value.executionTime = executionTime;
      return this;
    }

    /**
     * Set the properties from the given {@link Verifier}.
     *
     * @param verifier the {@link Verifier}
     * @return the current builder
     */
    public Builder fromVerifier(Verifier verifier) {
      return withVerificationType(Validate.notNull(verifier).getVerificationType());
    }

    @Override
    protected void validate() {
      super.validate();
      Validate.notNull(value.verificationType);
    }
  }
}
